@@ -61,6 +61,7 @@ S3_CN_NORTHWEST_HOST = 's3.cn-northwest-1.amazonaws.com.cn'
 S3_EU_WEST_HOST = 's3-eu-west-1.amazonaws.com'
 S3_EU_WEST2_HOST = 's3-eu-west-2.amazonaws.com'
 S3_EU_CENTRAL_HOST = 's3-eu-central-1.amazonaws.com'
+S3_EU_NORTH1_HOST = 's3-eu-north-1.amazonaws.com'
 S3_AP_SOUTH_HOST = 's3-ap-south-1.amazonaws.com'
 S3_AP_SOUTHEAST_HOST = 's3-ap-southeast-1.amazonaws.com'
 S3_AP_SOUTHEAST2_HOST = 's3-ap-southeast-2.amazonaws.com'
@@ -71,6 +72,32 @@ S3_SA_EAST_HOST = 's3-sa-east-1.amazonaws.com'
 S3_SA_SOUTHEAST2_HOST = 's3-sa-east-2.amazonaws.com'
 S3_CA_CENTRAL_HOST = 's3-ca-central-1.amazonaws.com'
 
+# Maps AWS region name to connection hostname
+REGION_TO_HOST_MAP = {
+    'us-east-1': S3_US_STANDARD_HOST,
+    'us-east-2': S3_US_EAST2_HOST,
+    'us-west-1': S3_US_WEST_HOST,
+    'us-west-2': S3_US_WEST_OREGON_HOST,
+    'us-gov-west-1': S3_US_GOV_WEST_HOST,
+    'cn-north-1': S3_CN_NORTH_HOST,
+    'cn-northwest-1': S3_CN_NORTHWEST_HOST,
+    'eu-west-1': S3_EU_WEST_HOST,
+    'eu-west-2': S3_EU_WEST2_HOST,
+    'eu-west-3': 's3.eu-west-3.amazonaws.com',
+    'eu-north-1': 's3.eu-north-1.amazonaws.com',
+    'eu-central-1': S3_EU_CENTRAL_HOST,
+    'ap-south-1': S3_AP_SOUTH_HOST,
+    'ap-southeast-1': S3_AP_SOUTHEAST_HOST,
+    'ap-southeast-2': S3_AP_SOUTHEAST2_HOST,
+    'ap-northeast-1': S3_AP_NORTHEAST1_HOST,
+    'ap-northeast-2': S3_AP_NORTHEAST2_HOST,
+    'ap-northeast-3': 's3.ap-northeast-3.amazonaws.com',
+    'sa-east-1': S3_SA_EAST_HOST,
+    'sa-east-2': S3_SA_SOUTHEAST2_HOST,
+    'ca-central-1': S3_CA_CENTRAL_HOST,
+    'me-south-1': 's3.me-south-1.amazonaws.com'
+}
+
 API_VERSION = '2006-03-01'
 NAMESPACE = 'http://s3.amazonaws.com/doc/%s/' % (API_VERSION)
 
@@ -85,7 +112,7 @@ RESPONSES_PER_REQUEST = 100
 class S3Response(AWSBaseResponse):
     namespace = None
     valid_response_codes = [httplib.NOT_FOUND, httplib.CONFLICT,
-                            httplib.BAD_REQUEST]
+                            httplib.BAD_REQUEST, httplib.PARTIAL_CONTENT]
 
     def success(self):
         i = int(self.status)
@@ -95,8 +122,12 @@ class S3Response(AWSBaseResponse):
         if self.status in [httplib.UNAUTHORIZED, httplib.FORBIDDEN]:
             raise InvalidCredsError(self.body)
         elif self.status == httplib.MOVED_PERMANENTLY:
-            raise LibcloudError('This bucket is located in a different ' +
-                                'region. Please use the correct driver.',
+            bucket_region = self.headers.get('x-amz-bucket-region', None)
+            used_region = self.connection.driver.region
+            raise LibcloudError('This bucket is located in a different '
+                                'region. Please use the correct driver. '
+                                'Bucket region "%s", used region "%s".' %
+                                (bucket_region, used_region),
                                 driver=S3StorageDriver)
         raise LibcloudError('Unknown error. Status code: %d' % (self.status),
                             driver=S3StorageDriver)
@@ -255,38 +286,29 @@ class BaseS3StorageDriver(StorageDriver):
         raise LibcloudError('Unexpected status code: %s' % (response.status),
                             driver=self)
 
-    def list_container_objects(self, container, ex_prefix=None):
-        """
-        Return a list of objects for the given container.
-
-        :param container: Container instance.
-        :type container: :class:`Container`
-
-        :param ex_prefix: Only return objects starting with ex_prefix
-        :type ex_prefix: ``str``
-
-        :return: A list of Object instances.
-        :rtype: ``list`` of :class:`Object`
-        """
-        return list(self.iterate_container_objects(container,
-                                                   ex_prefix=ex_prefix))
-
-    def iterate_container_objects(self, container, ex_prefix=None):
+    def iterate_container_objects(self, container, prefix=None,
+                                  ex_prefix=None):
         """
         Return a generator of objects for the given container.
 
         :param container: Container instance
         :type container: :class:`Container`
 
+        :param prefix: Only return objects starting with prefix
+        :type prefix: ``str``
+
         :param ex_prefix: Only return objects starting with ex_prefix
         :type ex_prefix: ``str``
 
         :return: A generator of Object instances.
         :rtype: ``generator`` of :class:`Object`
         """
+        prefix = self._normalize_prefix_argument(prefix, ex_prefix)
+
         params = {}
-        if ex_prefix:
-            params['prefix'] = ex_prefix
+
+        if prefix:
+            params['prefix'] = prefix
 
         last_key = None
         exhausted = False
@@ -447,8 +469,52 @@ class BaseS3StorageDriver(StorageDriver):
                              'chunk_size': chunk_size},
             success_status_code=httplib.OK)
 
+    def download_object_range(self, obj, destination_path, start_bytes,
+                              end_bytes=None, overwrite_existing=False,
+                              delete_on_failure=True):
+        self._validate_start_and_end_bytes(start_bytes=start_bytes,
+                                           end_bytes=end_bytes)
+
+        obj_path = self._get_object_path(obj.container, obj.name)
+
+        headers = {'Range': self._get_standard_range_str(start_bytes,
+                                                         end_bytes)}
+        response = self.connection.request(obj_path, method='GET',
+                                           headers=headers, raw=True)
+
+        return self._get_object(obj=obj, callback=self._save_object,
+                                response=response,
+                                callback_kwargs={
+                                    'obj': obj,
+                                    'response': response.response,
+                                    'destination_path': destination_path,
+                                    'overwrite_existing': overwrite_existing,
+                                    'delete_on_failure': delete_on_failure,
+                                    'partial_download': True},
+                                success_status_code=httplib.PARTIAL_CONTENT)
+
+    def download_object_range_as_stream(self, obj, start_bytes, end_bytes=None,
+                                        chunk_size=None):
+        self._validate_start_and_end_bytes(start_bytes=start_bytes,
+                                           end_bytes=end_bytes)
+
+        obj_path = self._get_object_path(obj.container, obj.name)
+
+        headers = {'Range': self._get_standard_range_str(start_bytes,
+                                                         end_bytes)}
+        response = self.connection.request(obj_path, method='GET',
+                                           headers=headers,
+                                           stream=True, raw=True)
+
+        return self._get_object(
+            obj=obj, callback=read_in_chunks,
+            response=response,
+            callback_kwargs={'iterator': response.iter_content(CHUNK_SIZE),
+                             'chunk_size': chunk_size},
+            success_status_code=httplib.PARTIAL_CONTENT)
+
     def upload_object(self, file_path, container, object_name, extra=None,
-                      verify_hash=True, ex_storage_class=None):
+                      verify_hash=True, headers=None, ex_storage_class=None):
         """
         @inherits: :class:`StorageDriver.upload_object`
 
@@ -458,6 +524,7 @@ class BaseS3StorageDriver(StorageDriver):
         return self._put_object(container=container, object_name=object_name,
                                 extra=extra, file_path=file_path,
                                 verify_hash=verify_hash,
+                                headers=headers,
                                 storage_class=ex_storage_class)
 
     def _initiate_multipart(self, container, object_name, headers=None):
@@ -640,7 +707,8 @@ class BaseS3StorageDriver(StorageDriver):
                                 (resp.status), driver=self)
 
     def upload_object_via_stream(self, iterator, container, object_name,
-                                 extra=None, ex_storage_class=None):
+                                 extra=None, headers=None,
+                                 ex_storage_class=None):
         """
         @inherits: :class:`StorageDriver.upload_object_via_stream`
 
@@ -660,10 +728,12 @@ class BaseS3StorageDriver(StorageDriver):
                                               extra=extra,
                                               stream=iterator,
                                               verify_hash=False,
+                                              headers=headers,
                                               storage_class=ex_storage_class)
         return self._put_object(container=container, object_name=object_name,
                                 extra=extra, method=method, query_args=params,
                                 stream=iterator, verify_hash=False,
+                                headers=headers,
                                 storage_class=ex_storage_class)
 
     def delete_object(self, obj):
@@ -782,8 +852,9 @@ class BaseS3StorageDriver(StorageDriver):
 
     def _put_object(self, container, object_name, method='PUT',
                     query_args=None, extra=None, file_path=None,
-                    stream=None, verify_hash=True, storage_class=None):
-        headers = {}
+                    stream=None, verify_hash=True, storage_class=None,
+                    headers=None):
+        headers = headers or {}
         extra = extra or {}
 
         headers.update(self._to_storage_class_headers(storage_class))
@@ -815,8 +886,18 @@ class BaseS3StorageDriver(StorageDriver):
         headers = response.headers
         response = response
         server_hash = headers.get('etag', '').replace('"', '')
+        server_side_encryption = headers.get('x-amz-server-side-encryption',
+                                             None)
+        aws_kms_encryption = (server_side_encryption == 'aws:kms')
+        hash_matches = (result_dict['data_hash'] == server_hash)
 
-        if (verify_hash and result_dict['data_hash'] != server_hash):
+        # NOTE: If AWS KMS server side encryption is enabled, ETag won't
+        # contain object MD5 digest so we skip the checksum check
+        # See https://docs.aws.amazon.com/AmazonS3/latest/API
+        # /RESTCommonResponseHeaders.html
+        # and https://github.com/apache/libcloud/issues/1401
+        # for details
+        if verify_hash and not aws_kms_encryption and not hash_matches:
             raise ObjectHashMismatchError(
                 value='MD5 hash {0} checksum does not match {1}'.format(
                     server_hash, result_dict['data_hash']),
@@ -834,7 +915,7 @@ class BaseS3StorageDriver(StorageDriver):
                 driver=self)
 
     def _put_object_multipart(self, container, object_name, stream,
-                              extra=None, verify_hash=False,
+                              extra=None, verify_hash=False, headers=None,
                               storage_class=None):
         """
         Uploads an object using the S3 multipart algorithm.
@@ -854,13 +935,16 @@ class BaseS3StorageDriver(StorageDriver):
         :keyword extra: Additional options
         :type extra: ``dict``
 
+        :keyword headers: Additional headers
+        :type headers: ``dict``
+
         :keyword storage_class: The name of the S3 object's storage class
         :type extra: ``str``
 
         :return: The uploaded object
         :rtype: :class:`Object`
         """
-        headers = {}
+        headers = headers or {}
         extra = extra or {}
 
         headers.update(self._to_storage_class_headers(storage_class))
@@ -869,12 +953,8 @@ class BaseS3StorageDriver(StorageDriver):
         meta_data = extra.get('meta_data', None)
         acl = extra.get('acl', None)
 
-        if not content_type:
-            content_type, _ = libcloud.utils.files.guess_file_mime_type(
-                object_name)
-
-        if content_type:
-            headers['Content-Type'] = content_type
+        headers['Content-Type'] = self._determine_content_type(
+            content_type, object_name)
 
         if meta_data:
             for key, value in list(meta_data.items()):
@@ -1001,10 +1081,37 @@ class BaseS3StorageDriver(StorageDriver):
 
 
 class S3StorageDriver(AWSDriver, BaseS3StorageDriver):
-    name = 'Amazon S3 (us-east-1)'
+    name = 'Amazon S3'
     connectionCls = S3SignatureV4Connection
     region_name = 'us-east-1'
 
+    def __init__(self, key, secret=None, secure=True, host=None, port=None,
+                 region=None, token=None, **kwargs):
+        # Here for backward compatibility for old and deprecated driver class
+        # per region approach
+        if hasattr(self, 'region_name') and not region:
+            region = self.region_name  # pylint: disable=no-member
+
+        self.region_name = region
+
+        if region and region not in REGION_TO_HOST_MAP.keys():
+            raise ValueError('Invalid or unsupported region: %s' % (region))
+
+        self.name = 'Amazon S3 (%s)' % (region)
+
+        if host is None:
+            host = REGION_TO_HOST_MAP[region]
+
+        super(S3StorageDriver, self).__init__(key=key, secret=secret,
+                                              secure=secure, host=host,
+                                              port=port,
+                                              region=region, token=token,
+                                              **kwargs)
+
+    @classmethod
+    def list_regions(self):
+        return REGION_TO_HOST_MAP.keys()
+
 
 class S3USEast2Connection(S3SignatureV4Connection):
     host = S3_US_EAST2_HOST
@@ -1109,6 +1216,17 @@ class S3APSEConnection(S3SignatureV4Connection):
     host = S3_AP_SOUTHEAST_HOST
 
 
+class S3EUNorth1Connection(S3SignatureV4Connection):
+    host = S3_EU_NORTH1_HOST
+
+
+class S3EUNorth1StorageDriver(S3StorageDriver):
+    name = 'Amazon S3 (eu-north-1)'
+    connectionCls = S3EUNorth1Connection
+    ex_location_name = 'eu-north-1'
+    region_name = 'eu-north-1'
+
+
 class S3APSEStorageDriver(S3StorageDriver):
     name = 'Amazon S3 (ap-southeast-1)'
     connectionCls = S3APSEConnection
